//
// Read directory from CP/M .dsk file.
//
// This is a stripped down version of a program in development that will
// allow the user to extract files from a .dsk image file and write them
// to the host filesystem. Works with raw disk images and JV3 images for
// CP/M only.
//
// 03/20/2005 - v1.0 First release.
// 03/21/2005 - v1.1	Fixed showDirectory to check side 1 (if it exists).
// 03/22/2005 - v1.2 Fixed bug that would skip directory sectors when 0xe5 was
//                   found in the user byte (could be erased file, not the end
//                   of the directory).
// 03/25/2005 - v1.3 Fixed incorrect computation of total of file sizes that
//                   would happen if there were erased files on the disk.
//                   Displays both active and erased files and amount of disk
//                   space used by both.
// 03/25/2005 - v1.4 Cleaned up code.
// 04/10/2005 - v1.5 Fixed incorrect computation of file size for files with
//                   more than one directory entry. Due to misunderstanding of
//                   how rc and ext in directory entry work. Fixed computation
//                   of format size for double sided disks.
//                   Added info, debug, type, and 1K options.
// 06/11/2005 - v1.6 Added option to search all .dsk files in current directory
//                   for wildcard file names.
/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *	GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<malloc.h>
#include	<string.h>
#include	<ctype.h>
#include	<dirent.h>
#include <sys/stat.h>

#include	"dskdir.h"
#include	"formats.h"

bool	ismatch(char *pattern, char *string);

//////////////////////
//
//      data
//
//////////////////////

int		directoryCount = 0;		// number of directory entries
DIRENTRY	directory[MAX_DIRECTORY_ENTRIES];

TRK_PTR	table = NULL;

char		formatDefs[] = "disk.fdf";
int		formatCount = 0;
int		builtinFormatCount = 0;
int		interleave[MAX_SPT];
FORMAT	*currentFormat = NULL;

char		*lenstr[4] = {
	"128", "256", "512", "1024"
};

bool	showTrackInfo = FALSE;
bool	debug = FALSE;
bool	sector0 = FALSE;	// will be true if raw format includes sector 0 (not CP/M!)
bool	jv3 = FALSE;		// true if JV3 format, false if raw track data
bool	noDir = FALSE;
bool	searchFlag = FALSE;
int	bfrlen = 0;
int	offset = 0;
int	disksize = 0;
int	diskoffset = 0;
int	spt = 0;
int	maxtrk = 0;
int	sectorsize = 0;
int	blocksize = 0;
int	recblock = 0;		// records per block
int	kilo = 1024;
int	currentTrk = 0;
int	currentSide = 0;
int	currentSec = 0;
int	directoryTrack;
FILE	*fp;
char	searchname[MAX_LEN];
struct stat		sbuf;
char	fname[256];
byte	sector[4];
byte	bfr[BFRSIZE];
byte	sectorbfr[SECSIZE];
byte	trackbfr[TRKSIZE];
byte	diskbfr[BFRSIZE];

// Get next byte from .dsk file.
// Return next byte or -1 if no more data.

int getNextByte(void)
{
	int	ret;

	if (!bfrlen || offset == bfrlen)
	{
		bfrlen = fread(bfr, sizeof(byte), BFRSIZE, fp);
		offset = 0;
	}

	if (bfrlen)
	{
		ret = (int) bfr[offset] & 0xff;
		offset++;
	}
	else
		ret = -1;

	return ret;
}

// Search for next sector in input.
// Return false if sector found, else return error.

bool findNextSector(void)
{
	int	data, i, len = 0, limit = MAX_SEARCH;
	bool	done = FALSE;
	bool	found = FALSE;
	bool	sync = FALSE;

	while (!done)
	{
		--limit;

		if (!limit)			// if we haven't found sync within max_search bytes,
		{
			found = FALSE;
			break;			// it isn't there--give up.
		}

		data = getNextByte();

		if (data < 0)
		{
			done = TRUE;
		}
		else if (data == SYNC)			// may be a sync string
		{
			sync = TRUE;

			for (i=0; i<8; i++)
			{
				data = getNextByte();

				if (data < 0)
				{
					done = TRUE;
					sync = FALSE;
					break;
				}

				if (data != SYNC)
				{
					sync = FALSE;
					break;
				}
			}

			if (sync)		// found a sync string, now search for ID address mark
			{
				while (!done)
				{
					data = getNextByte();

					if (data < 0)
						done = TRUE;
					else if (data == IAM)		// found ID address mark!
					{
						found = TRUE;
						done = TRUE;
						
						for (i=0; i<4; i++)
						{
							data = getNextByte();

							if (data < 0)
							{
								found = FALSE;
								done = TRUE;
							}
							else
								sector[i] = (byte) data;
						}

						if (found)				// now search for data address mark
						{
							done = FALSE;

							for (i=0; i<10; i++)		// skip some gap data
							{
								data = getNextByte();
								if (data < 0)
									done = TRUE;
							}

							while (!done && data >= 0)
							{
								data = getNextByte();

								if (data < 0)
								{
									found = FALSE;
									done = TRUE;
								}
								else if (data == DAM || data == ALTDAM)		// got the data address mark
								{
									found = TRUE;
									done = TRUE;

									switch (sector[3] & 3)	// now skip over the data
									{
										case 0:
											len = 128;
											break;

										case 1:
											len = 256;
											break;

										case 2:
											len = 512;
											break;

										case 3:
											len = 1024;
											break;
									}

									for (i=0; i<len; i++)
									{
										data = getNextByte();

										if (data < 0)
											break;

										sectorbfr[i] = (byte) data;
									}
								}
							}
						}
					}
				}
			}
		}
	}

	return !found;
}

// Move track data from input buffer to disk buffer.

void writeTrack(int size)
{
	int	i;

	for (i=0; i<size; i++)
	{
		diskbfr[diskoffset] = trackbfr[i];
		diskoffset++;
	}

	disksize += size;
}

// Skip to next number in string

char *nextNumber(char *cptr)
{
	while (*cptr)
	{
		if (*cptr == ',')
		{
			cptr++;
			break;
		}
		cptr++;
	}

	return cptr;
}

// Get predefined formats from 'disk.fdf' file.

int loadFormats(FILE *f)
{
	int		i, count = 0, offset = formatCount;
	FORMAT	*fmt;
	char		*cptr;

	while (!feof(f))
	{
		cptr = fgets(bfr, 256, f);

		if (!cptr)
			break;

		if (*cptr == '*')			// start of format definition
		{
			fmt = malloc(sizeof(FORMAT));
			cptr++;
			strcpy(fmt->name, cptr);
			i = strlen(cptr) - 1;

			while (cptr[i] == '\n' || cptr[i] == '\r')
			{
				--i;
				--cptr;
			}

			fmt->name[i] = '\0';
			cptr = fgets(bfr, 256, f);		// get dpb & option line

			if (!cptr)
				break;

			fmt->dpb.spt = (word) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.bsh = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.blm = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.exm = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.dsm = (word) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.drm = (word) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.al0 = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.al1 = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.cks = (word) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpb.off = (word) atoi(cptr);
			cptr = nextNumber(cptr);

			fmt->dpbext.pspt = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpbext.sizecode = (byte) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpbext.maxtrack = (word) atoi(cptr);
			cptr = nextNumber(cptr);
			fmt->dpbext.option = (byte) atoi(cptr);

			cptr = fgets(bfr, 256, f);		// get interleave line
			i = 0;

			if (!cptr)
				break;

			while (*cptr)
			{
				fmt->xlt[i] = (byte) atoi(cptr);
				i++;
				cptr = nextNumber(cptr);
			}

			Formats[offset] = fmt;
			offset++;
			count++;
		}
	}

	return count;
}

void freeFormats(void)
{
	int	i;

	for (i=builtinFormatCount; i<formatCount; i++)
		free(Formats[i]);
}

// Attempt to identify the disk format based on number of tracks, sectors,
// sector size, sides, and total disk size.
// If found among loaded formats, return a pointer to the matching format,
// else return NULL.

FORMAT *matchFormat(int trks, int secs, int size, int sides, int dsksize)
{
	int		i, sizec = 0, fmtsize, chksize, fmtsides;
	bool		found = FALSE;
	FORMAT	*fmt = NULL;

	switch (size)		// convert sector size to code
	{
		case 128:
			sizec = SS128;
			break;

		case 256:
			sizec = SS256;
			break;

		case 512:
			sizec = SS512;
			break;

		case 1024:
			sizec = SS1024;
			break;
	}

	for (i=0; i<formatCount; i++)
	{
		found = TRUE;
		fmtsides = (int) Formats[i]->dpbext.option & 0x40;

		if (fmtsides)
			fmtsides = 2;
		else
			fmtsides = 1;

		if (sides != fmtsides)
			found = FALSE;

		if (Formats[i]->dpbext.pspt != secs)
			found = FALSE;

		if (Formats[i]->dpbext.sizecode != sizec)
			found = FALSE;

		if (Formats[i]->dpbext.maxtrack != trks)
			found = FALSE;

		fmtsize = (int) Formats[i]->dpbext.pspt *
				((int) Formats[i]->dpbext.maxtrack -
				(int) Formats[i]->dpb.off) *
				size * fmtsides;

		if (fmtsides == 2)  // if double-sided, we subtracted too many sectors for track offset
			fmtsize += ((int) Formats[i]->dpb.off * (int) Formats[i]->dpbext.pspt * size);

		if (fmtsize != dsksize)
		{
			if (Formats[i]->dpb.off)
			{
				chksize = dsksize - (Formats[i]->dpb.off * Formats[i]->dpbext.pspt * size);

				if (fmtsize != chksize)
					found = FALSE;
			}
			else
				found = FALSE;
		}

		if (found)
		{
			fmt = Formats[i];
			break;
		}
	}

	return fmt;
}

// Try to identify directory sector

bool isDirectory(byte *dptr)
{
	int	i, off;
	bool	done = FALSE, ret = TRUE;	// must init or compiler complains

	for (i=0; i<4 && !done; i++)
	{
		ret = TRUE;
		dptr += (i * 32);

		if (dptr[0] < 16)
		{
			for (off=1; off<12; off++)		// check file name and extension
			{
				if (!isprint(dptr[off] & 0x7f))
				{
					ret = FALSE;
					done = TRUE;
					break;
				}
			}
		}

		if ((ret == TRUE) && (dptr[0] != 0xe5))	// filename checks ok, now look at binary data
		{
			for (off=16; off<31; off++)		// data map must be in increasing order (is this true?)
			{
				if (!dptr[off + 1])
					break;

				if (dptr[off] >= dptr[off + 1])
				{
					ret = FALSE;
					done = TRUE;
					break;
				}
			}
		}
	}

	return ret;
}

TRK_PTR getTrackPtr(int trk)
{
	return (TRK_PTR) table + trk;
}

SEC_PTR getSectorPtr(int trk, int side, int sec)
{
	TRK_PTR	tptr;
	SEC_PTR	sptr;

	if (!sector0)
		--sec;

	tptr = getTrackPtr(trk);
	sptr = tptr->side[side];
	sptr += sec;
	return sptr;
}

byte *getDataPtr(int trk, int side, int sec)
{
	SEC_PTR	sptr;
	byte		*ptr;

	sptr = getSectorPtr(trk, side, sec);
	ptr = sptr->dataptr;
	return ptr;
}

// Attempt to determine which track the directory is on.
// We assume that the directory will begin in sector 1.

int findDirTrack(void)
{
	int	trk, sec;
	byte	*dptr;

	if (sector0)
		sec = 0;
	else
		sec = 1;

	for (trk=0; trk<10; trk++)
	{
		dptr = getDataPtr(trk, 0, sec);

		if (dptr && isDirectory(dptr))
			return trk;
	}

	return -1;
}

// Search directory entries for filename,
// return index if found, or -1 if not found.

int getDirectoryIndex(char *name)
{
	int	i;

	for (i=0; i<directoryCount; i++)
	{
		if (!strcmp(name, directory[i].fname))
			return i;
	}

	return -1;
}

int addDirectoryEntry(char *name, FCB *fcb)
{
	int	i, cnt, index = directoryCount;

	strcpy(directory[directoryCount].fname, name);
	directory[directoryCount].size = 0;
	directory[directoryCount].fcb = malloc(sizeof(FCBENTRY));
	directory[directoryCount].fcb->next = NULL;
	memcpy(directory[directoryCount].fcb, fcb, sizeof(FCB));
	cnt = 0;

	for (i=0; i<16; i++)
	{
		if (fcb->map[i])
			cnt++;
	}

	if (fcb->reccnt == 128)
		directory[directoryCount].size = (cnt * blocksize);
	else if (cnt > 0)
	{
		--cnt;
		directory[directoryCount].size = (cnt * blocksize);
		directory[directoryCount].size += ((fcb->reccnt % recblock) * 128);
	}
	else
		directory[directoryCount].size = 0;

	directoryCount++;
	return index;
}

void freeDirectoryEntries(void)
{
	int		i;
	FCBENTRY	*fcb, *next;

	for (i=0; i<directoryCount; i++)
	{
		fcb = directory[i].fcb;

		while (fcb)
		{
			next = fcb->next;
			free(fcb);
			fcb = next;
		}
	}

	directoryCount = 0;
}

void addFCB(DIRENTRY *dir, FCB *fcb)
{
	int		i, cnt;
	FCBENTRY	*next;

	next = dir->fcb;

	while (next->next)
		next = next->next;

	next->next = malloc(sizeof(FCBENTRY));
	next = next->next;
	next->next = NULL;
	memcpy(next, fcb, sizeof(FCB));
	cnt = 0;

	for (i=0; i<16; i++)
	{
		if (fcb->map[i])
			cnt++;
	}

	if (fcb->reccnt == 128)
		dir->size += (cnt * blocksize);
	else
	{
		--cnt;
		dir->size += (cnt * blocksize);
		dir->size += ((fcb->reccnt % recblock) * 128);
	}
}

// Make reasonable guess as to whether the string is a file name

bool isFileName(char *str)
{
	int	i;

	for (i=0; i<15; i++)
	{
		if (!isprint(str[i]))
			return FALSE;
	}

	return TRUE;
}

// Build file list (into directory[]) from directory sectors

int getFiles(void)
{
	int	i, j, off, len, num, size, idx, total, etotal;
	int	dtrk, dside, dsec;
	int	files, efiles;
	byte	*ptr;
	FCB	*fcb;
	char	chr, name[16];

	dtrk = directoryTrack;
	num = 0;
	total = etotal = 0;
	files = efiles = 0;
	directoryCount = 0;

	while (dtrk < directoryTrack + 2)
	{
		dsec = 1;
		dside = 0;

		while (dside < 2)
		{
			while (dsec < spt)
			{
				ptr = getDataPtr(dtrk, dside, dsec);

				if (ptr && isDirectory(ptr))
				{
					for (i=0; i<sectorsize; )
					{
						if ((ptr[i] == 0xe5) && (ptr[i + 1] == 0xe5))		// end of directory entries
							break;

						if (ptr[i] == 0xe5)
							name[0] = '*';
						else
							name[0] = ' ';

						len = 1;

						for (j=1, off = i + 1; j<12; j++)
						{
							if (j == 9)
							{
								name[len] = '.';
								len++;
							}

							chr = (char) ptr[off++] & 0x7f;

							if (chr != ' ')
							{
								name[len] = chr;
								len++;
							}
						}

						while (len < 15)
						{
							name[len] = ' ';
							len++;
						}

						name[len] = '\0';

						if (isFileName(name))
						{
							size = (int) ptr[i + 15] & 0xff;		// record count

							if (size == 128)
								size *= 256;
							else
								size *= 128;

							total += size;
							fcb = (FCB *) &ptr[i];

							idx = getDirectoryIndex(name);

							if (idx < 0)
							{
								if (name[0] == '*')
									efiles++;
								else
									files++;

								idx = addDirectoryEntry(name, fcb);
							}
							else
								addFCB((DIRENTRY *) &directory[idx], fcb);
						}

						i += 32;
					}
				}
				dsec++;
			}
			dside++;
		}
		dtrk++;
	}

	return files + efiles;
}

// Display files in directory

int showDirectory()
{
	int	i, j, num, size, bsize, total, etotal;
	int	dtrk;
	int	files, efiles;
	char	*cptr;
	FCBENTRY	*fcb;
	char	cbfr[16];

	dtrk = directoryTrack;
	num = 0;
	total = etotal = 0;
	files = efiles = 0;

	directoryCount = getFiles();

	if (searchFlag)
	{
		for (i=0; i<directoryCount; i++)
		{
			cptr = (char *) &directory[i].fname[1];

			if (ismatch(searchname, cptr))
			{
				j = printf("\n%s  ", fname);

				while (j < 24)
				{
					printf(" ");
					j++;
				}

				printf("%s", cptr);
			}
		}

		return 0;
	}

	printf("\n\n");

	for (i=0; i<directoryCount; i++)
	{
		printf("%s", directory[i].fname);
		size = directory[i].size;

		if (directory[i].fname[0] == '*')
		{
			efiles++;
			etotal += size;
		}
		else
		{
			files++;
			total += size;
		}

		bsize = size;

		if (size < (10 * kilo))
			sprintf(cbfr, "%5d", size);
		else
		{
			size /= kilo;							// file size in K bytes
			sprintf(cbfr, "%4dK", size);
		}

		printf("%s  ", cbfr);
		num++;

		if (debug)
			printf("%d bytes  ", bsize);

		if ((!debug) && (num > 3))
		{
			num = 0;
			printf("\n");
		}

		if (debug)
		{
			printf("\n");
			fcb = directory[i].fcb;

			while (fcb)
			{
				printf("     ");
				cptr = (char *) fcb;

				for (j=1; j<12; j++)
						printf("%c", cptr[j] & 0xff);

				printf(", ext %d, s1 %d, s2 %d, rec %d, map:\n    ",
					cptr[12] & 0xff, cptr[13] & 0xff, cptr[14] & 0xff, cptr[15] & 0xff);

				for (j=16; j<32; j++)
					printf(" %02x", cptr[j] & 0xff);

				printf("\n");
				fcb = fcb->next;
			}
		}
	}

	if ((!debug) && num)
		printf("\n");

	printf("\nTotal %d files, ", files);

	if (total < (10 * kilo))
		printf("%d bytes", total);
	else
	{
		total /= kilo;
		printf("%d Kbytes", total);
	}

	if (efiles)
	{
		printf(", %d erased files with ", efiles);

		if (etotal < (10 * kilo))
			printf("%d bytes", etotal);
		else
		{
			etotal /= kilo;
			printf("%d Kbytes", etotal);
		}

		printf("\n(Erased files, marked with '*', may not be recoverable)");
	}

	return total;
}

void initTrackTable(void)
{
	int		i, j;
	TRK_PTR	tptr;
	SEC_PTR	sptr;

	table = malloc(sizeof(TRACK_ENTRY) * MAX_TRACK);

	for (i=0; i<MAX_TRACK; i++)
	{
		tptr = table + i;
		tptr->spt = 0;

		for (j=0; j<MAX_SPT; j++)
			tptr->xlt[j] = 0;

		tptr->side[0] = malloc(MAX_SPT * sizeof(SEC_ENTRY));
		sptr = tptr->side[0];

		for (j=0; j<MAX_SPT; j++)
		{
			sptr->size = 0;
			sptr->dataptr = NULL;
			sptr++;
		}

		tptr->side[1] = malloc(MAX_SPT * sizeof(SEC_ENTRY));
		sptr = tptr->side[1];

		for (j=0; j<MAX_SPT; j++)
		{
			sptr->size = 0;
			sptr->dataptr = NULL;
			sptr++;
		}
	}
}

void deleteTrackTable(void)
{
	int		i;
	TRK_PTR	tptr;

	if (!table)
		return;

	for (i=0; i<MAX_TRACK; i++)
	{
		tptr = table + i;

		if (tptr)
		{
			if (tptr->side[0])
				free(tptr->side[0]);

			if (tptr->side[1])
				free(tptr->side[1]);
		}
	}

	free(table);
	table = NULL;
}

//
// Scan the .dsk file and attempt to identify the disk image type.
//
// Raw format refers to a .dsk that is a byte image of the disk, including
// FDC sync data. JV3 dsk files have a header of 0x2200 (8704) bytes that
// describe the layout in terms of 3 byte sets, where the first byte is the
// track number, the second byte is the sector number, and the third byte
// bit 4 is the side. JV3 dsk files have a sector size of 256 bytes. The
// last byte of the header for JV3 is a write protect flag.
//

void scanDisk(FILE *fp)
{
	int	i, j, len = 0;
	int	sec = 0, maxsec = 0, secsize = 0;
	int	trk = 0, lasttrk = -1, lastside = -1;
	int	side = 0, sides = 0;
	int	xltoffset = 0;
	int	trkoffset = 0;
	bool	done = FALSE;
	TRK_PTR	tptr = NULL;
	SEC_PTR	sptr = NULL;

	while (!done)		// scan the dsk file, assuming first that it's raw format
	{
		done = findNextSector();

		if (!done)
		{
			if (trk != (int) sector[0] || side != (int) sector[1])	// if switched side or track
			{
				if (showTrackInfo)
				{
					printf("\nTrack %2d, side %d: ", trk, side);

					for (j=0; j<xltoffset; j++)
					{
						if (sector0)
							printf(" %2d", interleave[j] - 1);
						else
							printf(" %2d", interleave[j]);
					}
				}

				writeTrack(trkoffset);	// dump track buffer to disk buffer

				for (i=0; i<xltoffset; i++)
					tptr->xlt[i] = interleave[i];

				tptr->xlt[xltoffset] = 0;
				tptr->spt = xltoffset;
				trkoffset = 0;
				xltoffset = 0;
			}

			trk = (int) sector[0];			// move sector buffer to track buffer
			side = (int) sector[1];
			sec = (int) sector[2];

			if (!sec)
				sector0 = TRUE;

			lasttrk = trk;
			tptr = getTrackPtr(trk);
			tptr->xlt[xltoffset] = sec;

			if (sector0)
				interleave[xltoffset] = sec + 1;
			else
				interleave[xltoffset] = sec;

			xltoffset++;

			if (trk > maxtrk)
				maxtrk = trk;

			if (side > sides)
				sides = side;

			if (sec > maxsec)
				maxsec = sec;

			switch (sector[3] & 3)
			{
				case 0:
					len = 128;
					break;

				case 1:
					len = 256;
					break;

				case 2:
					len = 512;
					break;

				case 3:
					len = 1024;
					break;
			}

			sptr = getSectorPtr(trk, side, sec);
			sptr->size = len;
			sptr->dataptr = diskbfr + diskoffset + (sec - 1) * len;

			if (len > secsize)
				secsize = len;

			for (i=0; i<len; i++)
			{
				trackbfr[trkoffset] = sectorbfr[i];
				trkoffset++;
			}
		}
	}

	if (trkoffset)
	{
		writeTrack(trkoffset);

		for (i=0; i<xltoffset; i++)
			tptr->xlt[i] = interleave[i];

		tptr->xlt[xltoffset] = 0;
		tptr->spt = xltoffset;

		if (showTrackInfo)
		{
			printf("\nTrack %2d, side %d: ", trk, side);
			for (j=0; j<xltoffset; j++)
			{
				if (sector0)
					printf(" %2d", interleave[j] - 1);
				else
					printf(" %2d", interleave[j]);
			}
		}
	}

	fclose(fp);

	if (showTrackInfo)
		printf("\n");

	if (!disksize)		// if disk size is non-zero, test if it's JV3 format
	{
		lasttrk = -1;
		fp = fopen(fname, "rb");
		bfrlen = fread(bfr, sizeof(byte), 0x2200, fp);			// read header into bfr
		disksize = fread(diskbfr, sizeof(byte), BFRSIZE, fp);	// read data into diskbfr
		fclose(fp);
		xltoffset = 0;
		j = 0;

		for (i=0; i<0x2200; )
		{
			trk = (int) bfr[i++] & 0xff;
			sec = (int) bfr[i++] & 0xff;
			side = (int) bfr[i++] & 0x10;

			if (!sec)
				sector0 = TRUE;

			if (side)
				side = 1;
			else
				side = 0;

			if (trk > 80)		// end of header
				break;

			if (side > sides)
				sides = side;

			if ((trk != lasttrk) || (side != lastside))
			{
				lasttrk = trk;
				lastside = side;
				xltoffset = j = 0;

				if (trk > maxtrk)
					maxtrk = trk;
			}

			interleave[j] = sec;
			j++;

			if (sec > maxsec)
				maxsec = sec;

			if (j > xltoffset)
				xltoffset = j;
		}

		secsize = disksize / ((maxtrk + 1) * maxsec * (sides + 1));
		xltoffset = 0;
		j = 0;
		jv3 = TRUE;
		len = 256;	// JV3 dsk files always have a length of 256 bytes/sector

		for (i=0; i<0x2200; )		// now rescan header and set up track/sector data pointers
		{
			trk = (int) bfr[i++] & 0xff;
			sec = (int) bfr[i++] & 0xff;
			side = (int) bfr[i++] & 0x10;

			if (side)
				side = 1;
			else
				side = 0;

			if (trk > 80)		// end of header
				break;

			if ((trk != lasttrk) || (side != lastside))
			{
				tptr = getTrackPtr(lasttrk);
				tptr->spt = j;

				for (j=0; j<xltoffset; j++)
					tptr->xlt[j] = interleave[j];

				tptr->xlt[j] = 0;

				lasttrk = trk;
				lastside = side;
				xltoffset = j = 0;
			}

			interleave[j] = sec;
			j++;

			if (j > xltoffset)
				xltoffset = j;

			sptr = getSectorPtr(trk, side, sec);
			sptr->size = len;
			sptr->dataptr = diskbfr + diskoffset;
			diskoffset += len;
		}

		tptr = getTrackPtr(lasttrk);
		tptr->spt = j;

		for (j=0; j<xltoffset; j++)
			tptr->xlt[j] = interleave[j];

		tptr->xlt[j] = 0;
	}		// JV3 format

	sectorsize = secsize;
	spt = maxsec;

	directoryTrack = findDirTrack();

	if (directoryTrack < 0)
		directoryTrack = 0;

	currentFormat = matchFormat(maxtrk + 1, maxsec, secsize, sides + 1, disksize);

	if (currentFormat)
	{
		directoryTrack = (int) currentFormat->dpb.off / (sides + 1);

		if (!searchFlag)
			printf("\n%s appears to be \"%s\"", fname, currentFormat->name);
	}
	else		// Doesn't match any known format, attempt to construct
	{			// a reasonable dpb from disk parameters. NOTE: these things
				// always seem to store the sectors in order, regardless of
				// the actual interleave on the disk.
		if (!searchFlag)
			printf("\n%s - unknown format", fname);

		currentFormat = malloc(sizeof(FORMAT));
		strcpy(currentFormat->name, "Unknown format");
		currentFormat->dpb.spt = spt;
		currentFormat->dpb.bsh = 4;
		currentFormat->dpb.blm = 0;
		currentFormat->dpb.exm = 0;
		currentFormat->dpb.dsm = 0;
		currentFormat->dpb.drm = 0;
		currentFormat->dpb.al0 = 0;
		currentFormat->dpb.al1 = 0;
		currentFormat->dpb.off = 2;				// offset may not be 2 for all JV3 images
		currentFormat->dpbext.pspt = maxsec;
		currentFormat->dpbext.sizecode = SS256;
		currentFormat->dpbext.maxtrack = maxtrk + 1;

		if (sides)
			currentFormat->dpbext.option = 0x40;
		else
			currentFormat->dpbext.option = 0;

		currentFormat->dpbext.type = 1;

		for (i=0; i<xltoffset; i++)
			currentFormat->xlt[i] = interleave[i];
	}

	blocksize = 128 << currentFormat->dpb.bsh;
	recblock = blocksize / 128;

	if (debug)
		printf("\nblock size = %d, records/block = %d\n", blocksize, recblock);
}

//
// Case insensitive match of string against pattern.
//
// Wildcard rules:
//
// 1) '?' in pattern matches any single character.
//
// 2) '*' in pattern matches any sequence of characters,
//    or none at all, until string character matches
//    the following character in pattern (if any).
//

bool ismatch(char *pattern, char *string)
{
	int	i, j, wc_index;
	bool	ret;
	char	next;
	char	str[MAX_LEN];
	char	pat[MAX_LEN];

//printf("Matching '%s' to '%s'\n", pattern, string);

	ret = FALSE;
	wc_index = -1;
	i = 0;

	while (pattern[i])					// convert pattern to upper case
	{
		pat[i] = toupper(pattern[i]);
		i++;
	}

	pat[i] = '\0';
	i = 0;

	while (string[i])						// convert string to upper case
	{
		str[i] = toupper(string[i]);
		i++;
	}

	str[i] = '\0';

	while (--i)
	{
		if (str[i] == ' ')
			str[i] = '\0';
		else
			break;
	}

	i = j = 0;

	while (1)
	{
		if (!pat[i] && !str[j])			// both empty, found a match
		{
			ret = TRUE;
			break;
		}

		if (!pat[i] && str[j])			// pattern empty, but string isn't
			break;							// no match

		if (!str[j])						// string is empty
		{
			next = pat[i + 1];			// check next in pattern

			if (!next && (pat[i] == '?' || pat[i] == '*'))
				ret = TRUE;					// match if next is wildcard

			break;
		}

		if (pat[i] == '?')				// match any single character
		{
			i++;
			j++;
		}
		else if (pat[i] == '*')			// match any sequence of chars
		{
			wc_index = i;					// save current index of wildcard
			next = pat[i + 1];			// get next pattern character

			while (str[j] && str[j] != next)
				j++;							// go through string til next match

			i++;								// then increment pattern index
		}
		else									// match next character
		{
			if (pat[i] != str[j])		// no match
			{
				if (wc_index >= 0)		// if we had a wildcard,
					i = wc_index - 1;		// back up to it
				else							// else match failed
					break;
			}
			i++;								// next pattern character
			j++;								// next string character
		}
	}

	return ret;
}

char *findfirst(char *ftemplate, struct dirent *dp, DIR *dirp)
{
	char				*fname;
	struct dirent	*localdp;

	strcpy(searchname, ftemplate);

	while ((localdp = readdir(dirp)) != NULL)
	{
		lstat(localdp->d_name, &sbuf);

		if ((sbuf.st_mode & S_IFMT) != S_IFDIR)
		{
			if (ismatch(searchname, localdp->d_name))
			{
				fname = malloc(strlen(localdp->d_name) + 1);
				strcpy(fname, localdp->d_name);
				return fname;
			}
		}
	}

	return NULL;
}

char *findnext(struct dirent *dp, DIR *dirp)
{
	char				*fname;
	struct dirent	*localdp;

	while ((localdp = readdir(dirp)) != NULL)
	{
		lstat(localdp->d_name, &sbuf);

		if ((sbuf.st_mode & S_IFMT) != S_IFDIR)
		{
			if (ismatch(searchname, localdp->d_name))
			{
				fname = malloc(strlen(localdp->d_name) + 1);
				strcpy(fname, localdp->d_name);
				return fname;
			}
		}
	}

	return NULL;
}

void doDisk()
{
	int	i;

	fp = fopen(fname, "rb");

	if (!fp)
	{
		printf("\nInput file %s not found\n\n", fname);
		return;
	}

	initTrackTable();
	scanDisk(fp);

	if (!noDir)
	{

// at this point diskbfr contains the actual sector data
// without the raw header info.

		i = showDirectory();

		if (!i && !searchFlag)
			printf("\nThis disk may not be a CP/M disk image");
	}
	else if (!searchFlag)
	{
		if (jv3)
			printf("\nJV3 format");
		else
			printf("\nRaw format");
	}
}

void usage(void)
{
	printf("\ndskdir version %d.%d", VERSION, REVISION);
	printf("\nRead directory information from CP/M .dsk file\n");
	printf("\nUsage: dskdir [-d|h|i|k|t|?] dskFileName");
	printf("\nOptions: -d output debug info.");
	printf("\n         -h or -? show this usage message.");
	printf("\n         -i output track/sector info.");
	printf("\n         -k use 1000 to compute file size in KB (default = 1024).");
	printf("\n         -s search for CP/M files in .dsk files.");
	printf("\n            Example: dskdir -s file.com, searches all .dsk files for 'file.com'.");
	printf("\n            If filename is wildcarded, it must be enclosed in single quotes.");
	printf("\n            Example: dskdir -s '*.txt'.");
	printf("\n         -t show disk format (type) only (no directory).");
	printf("\n\n");
	exit(0);
}

int main(int argc, char *argv[])
{
	int	i;
	FILE	*fmtfd;
	char	*cptr;
	DIR	*dirp;
	struct dirent	finfo;
	FILELIST	*dskfiles, *dsknext;

	if (argc < 2)
		usage();

	for (i=1; i<argc; i++)
	{
		cptr = argv[i];

		if (*cptr == '-')
		{
			switch (toupper(cptr[1]))
			{
				case 'D':
					debug = TRUE;
					break;

				case 'I':
					showTrackInfo = TRUE;
					break;

				case 'K':
					kilo = 1000;
					break;

				case 'S':
					searchFlag = TRUE;
					break;

				case 'T':
					noDir = TRUE;
					break;

				case '?':
				case 'H':
					usage();
					break;
			}
		}
		else
			strcpy(fname, argv[i]);
	}

	for (i=0; ; i++)
	{
		if (Formats[i])
			formatCount++;
		else
			break;
	}

	builtinFormatCount = formatCount;

	fmtfd = fopen(formatDefs, "r");

	if (fmtfd)
	{
		i = loadFormats(fmtfd);
		fclose(fmtfd);
		formatCount += i;
	}

	if (!searchFlag)
	{
		cptr = strrchr(fname, '.');

		if (!cptr)
			strcat(fname, ".dsk");

		doDisk();
	}
	else		// search for specific files in all .dsk files
	{
		dskfiles = NULL;
		dirp = opendir(".");			// get pointer to directory stream

		cptr = findfirst("*.dsk", (struct dirent *) &finfo, dirp);

		if (!cptr)
		{
			printf("\nNo file found\n");
			return 1;
		}

		dskfiles = malloc(sizeof(FILELIST));
		dskfiles->fname = cptr;
		dskfiles->next = NULL;
		dsknext = dskfiles;

		while (1)
		{
			cptr = findnext((struct dirent *) &finfo, dirp);

			if (!cptr)
				break;

			dsknext->next = malloc(sizeof(FILELIST));
			dsknext->next->fname = cptr;
			dsknext->next->next = NULL;
			dsknext = dsknext->next;
		}

		closedir(dirp);		// done with this directory
		strcpy(searchname, fname);
		dsknext = dskfiles;

		while (dsknext)		// Now check all the files we found...
		{
			strcpy(fname, dsknext->fname);
			doDisk(0);
			dsknext = dsknext->next;
			freeDirectoryEntries();

			if (table)
				deleteTrackTable();

			bfrlen = 0;
			offset = 0;
			disksize = 0;
			diskoffset = 0;
			spt = 0;
			maxtrk = 0;
			sectorsize = 0;
			blocksize = 0;
			recblock = 0;
			currentTrk = 0;
			currentSide = 0;
			currentSec = 0;
			directoryCount = 0;
		}

		while (dskfiles)
		{
			dsknext = dskfiles->next;
			free(dskfiles->fname);
			free(dskfiles);
			dskfiles = dsknext;
		}
	}

	printf("\n\n");

	if (table)
		deleteTrackTable();

	freeFormats();
	freeDirectoryEntries();
	return 0;
}

// end

